from mongoframes import *

from manhattan.comparable.frames import ComparableFrame

__all__ = [
    'Dragon',
    'Lair',
    'User'
    ]


class Lair(ComparableFrame):
    """
    A lair in which a dragon resides.
    """

    _fields = {
        'name'
        }

    def __str__(self):
        return self.name


class Dragon(ComparableFrame):
    """
    A dragon.
    """

    _fields = {
        'name',
        'breed',
        'hobbies',
        'lair',
        'secret_password'
        }

    _compared_refs = {'lair': Lair}

    _default_projection = {
        'lair': {'$ref': Lair}
        }

    def __str__(self):
        return self.name


class User(ComparableFrame):
    """
    A user, a keeper of records.
    """

    _fields = {'name'}

    def __str__(self):
        return self.name


from mongoframes import *
from pymongo import MongoClient
import pytest

from manhattan.comparable.frames import ComparableFrame

from . import Dragon, Lair, User


# Fixtures

@pytest.fixture(scope='function')
def mongo_client(request):
    """Connect to the test database"""

    # Connect to mongodb and create a test database
    db_name = 'manhattan_comparable_test'
    Frame._client = MongoClient(
        'mongodb://localhost:27017/' + db_name)

    def fin():
        # Remove the test database
        Frame._client.drop_database('manhattan_comparable_test')

    request.addfinalizer(fin)

    return Frame._client

@pytest.fixture(scope='function')
def dragon(request):
    """Create a dragon"""

    dungeon = Lair(name='Dungeon')
    dungeon.insert()

    burt = Dragon(
        name='Burt',
        breed='Cold-drake',
        lair=dungeon,
        hobbies=['arson', 'death from above']
        )
    burt.insert()

    return burt

@pytest.fixture(scope='function')
def lair(request):
    """Create a lair"""
    cave = Lair(name='Cave')
    cave.insert()

    return cave

@pytest.fixture(scope='function')
def user(request):
    """Create a user"""
    fred = User(name='Fred')
    fred.insert()

    return fred


from mongoframes import *
from datetime import date

__all__ = ['ChangeLogEntry']


class ChangeLogEntry(Frame):
    """
    The `ComparableFrame` class tracks changes to documents through entries in a
    change log. The `ChangeLogEntry` class is provided as a base class for
    implementing a change log but can also be used directly.
    """

    _fields = {
        'created',
        'documents',
        'documents_sticky_label',
        'user',
        'user_sticky_label',
        'type',
        'details'
        }

    _indexes = [
        IndexModel([('documents', ASC)])
    ]

    @property
    def diff_html(self):
        """Return the entry's diff in HTML format"""
        return self.diff_to_html(self.details)

    @property
    def is_diff(self):
        """Return True if there are any differences logged for the entry"""
        if not isinstance(self.details, dict):
            return False

        for key in ['additions', 'updates', 'deletions']:
            if self.details.get(key, None):
                return True

        return False

    def add_diff(self, original, new):
        """
        Set the details of the change log entry as the difference between two
        dictionaries (original vs. new). The change log uses the following
        format:

            {
                'additions': {
                    'field_name': 'value',
                    ...
                },
                'updates': {
                   'field_name': ['original_value', 'new_value'],
                    ...
                },
                'deletions': {
                    'field_name': ['original_value']
                }
            }

        Values are tested for equality, there is special case handling for
        `Frame` class instances (see `diff_safe`) and fields with the word
        password in their name are redacted.

        NOTE: Where possible use diff structures that are flat, performing a
        diff on a dictionary which contains sub-dictionaries is not recommended
        as the verbose output (see `diff_to_html`) is optimized for flat
        structures.
        """
        changes = {}

        # Check for additions and updates
        for new_key, new_value in new.items():

            # Additions
            if new_key not in original:
                if 'additions' not in changes:
                    changes['additions'] = {}
                new_value = self.diff_safe(new_value)
                changes['additions'][new_key] = new_value

            # Updates
            elif not self.cmp(original[new_key], new_value):
                if 'updates' not in changes:
                    changes['updates'] = {}

                original_value = self.diff_safe(original[new_key])
                new_value = self.diff_safe(new_value)

                changes['updates'][new_key] = [original_value, new_value]

                # Check for password type fields and redact them
                if 'password' in new_key:
                    changes['updates'][new_key] = ['*****', '*****']

        # Check for deletions
        for original_key, original_value in original.items():
            if original_key not in new:
                if 'deletions' not in changes:
                    changes['deletions'] = {}

                original_value = self.diff_safe(original_value)
                changes['deletions'][original_key] = original_value

        self.details = changes

    def add_note(self, note):
        """Add a note as the entries details"""
        self.details = {'note': note}

    @classmethod
    def cmp(cls, value_a, value_b):
        """Return True if 2 values are the same"""

        if type(value_a) is not type(value_b):
            return False

        if isinstance(value_a, (list, tuple)):
            if len(value_a) != len(value_b):
                return False

            for i, a in enumerate(value_a):
                b = value_b[i]
                if not cls.cmp(a, b):
                    return False

            return True

        if isinstance(value_a, dict):
            for k, v in value_a.items():
                if not cls.cmp(v, value_b.get(k)):
                    return False

            return True

        return value_a == value_b

    @classmethod
    def diff_safe(cls, value):
        """Return a value that can be safely stored as a diff"""
        if isinstance(value, Frame):
            return {'_str': str(value), '_id': value._id}
        elif isinstance(value, (list, tuple)):
            return [cls.diff_safe(v) for v in value]
        return value

    @staticmethod
    def _on_insert(sender, frames):
        for frame in frames:

            # Record *sticky* labels for the change so even if the documents or
            # user are removed from the system their details are retained.
            pairs = [(d, d.__class__.__name__) for d in frame.documents]
            frame.documents_sticky_label = ', '.join(
                ['{0} ({1})'.format(*p) for p in pairs]
                )

            if frame.user:
                frame.user_sticky_label = str(frame.user)


ChangeLogEntry.listen('insert', ChangeLogEntry.timestamp_insert)
ChangeLogEntry.listen('insert', ChangeLogEntry._on_insert)



from datetime import date
import pytest

from manhattan.comparable.change_log import ChangeLogEntry
from manhattan.comparable.frames import ComparableFrame

from . import Dragon, Lair, User

pytest_plugins = ['manhattan.comparable.tests.fixtures']


# Tests

def test_init(mongo_client):
    """Should create a new Dragon instance"""

    # Passing no inital values
    burt = Dragon()
    assert isinstance(burt, Dragon)

    # Check the '_id', 'created' and 'modified' fields have been added by
    # default.
    assert '_id' in burt._fields
    assert 'created' in burt._fields
    assert 'modified' in burt._fields

    # Check the collection name has been set as the class name
    assert burt._collection == 'Dragon'

    # Passing initial values
    burt = Dragon(
        name='Burt',
        breed='Cold-drake'
        )
    assert burt.name == 'Burt'
    assert burt.breed == 'Cold-drake'

def test_comparable(mongo_client, dragon):
    """Should return a dictionary that can be compared"""
    assert dragon.comparable == {
        'breed': 'Cold-drake',
        'hobbies': ['arson', 'death from above'],
        'lair': Lair.by_id(dragon.lair._id),
        'name': 'Burt'
        }

def test_logged_delete(mongo_client, dragon, user):
    """Should delete the document and log the event in the change log"""
    dragon.logged_delete(user)

    # Check the document was deleted
    assert Dragon.by_id(dragon._id) is None

    # Check an entry for the deletion was stored
    entry = ChangeLogEntry.one()
    assert entry.type == 'DELETED'
    assert entry.documents == [dragon._id]
    assert entry.user == user._id

def test_logged_insert(mongo_client, user):
    """
    Should create and insert the document and log the event in the change log.
    """

    burt = Dragon(
        name='Burt',
        breed='Cold-drake'
        )
    burt.logged_insert(user)

    # Check the dragon was created
    assert burt._id is not None

    # Check an entry for the insert was stored
    entry = ChangeLogEntry.one()
    assert entry.type == 'ADDED'
    assert entry.documents == [burt._id]
    assert entry.user == user._id

def test_logged_update(mongo_client, dragon, user):
    """
    Should update the document with the dictionary dat provided and log the
    event in the change log.
    """
    dragon.logged_update(user, {'name': 'Burtress'})

    # Check the dragon was updated
    dragon.reload()
    assert dragon.name == 'Burtress'

    # Check an entry for the update was stored
    entry = ChangeLogEntry.one()
    assert entry.type == 'UPDATED'
    assert entry.documents == [dragon._id]
    assert entry.user == user._id
    assert entry.details == {'updates': {'name': ['Burt', 'Burtress']}}

def test_logged_update_no_update(mongo_client, dragon, user):
    """Should log no event if no comparable change was detected"""
    dragon.logged_update(user, {'name': 'Burt'})

    # Check no entry was created
    entry = ChangeLogEntry.one()
    assert entry is None

def test_compare_safe():
    """
    Should return a value that can safely be compared, by default this method
    simply converts dates to strings.
    """

    today = date.today()

    unsafe = {
        'foo': 'bar',
        'date': today,
        'dates': [today],
        'date_table': {'today': today}
    }

    safe = {
        'foo': 'bar',
        'date': str(today),
        'dates': [str(today)],
        'date_table': {'today': str(today)}
    }

    assert ComparableFrame.compare_safe(unsafe) == safe


from mongoframes import *
from mongoframes.frames import _FrameMeta
from datetime import date

from manhattan.comparable.change_log import ChangeLogEntry

__all__ = ['ComparableFrame']


class _ComparableFrameMeta(_FrameMeta):
    """
    Meta class for `ComparableFrame`s to ensure that `created` and `modified`
    are present in any defined set of fields (along with `_id`).
    """

    def __new__(meta, name, bases, dct):

        # If a set of fields is defined ensure it contains `created` and
        # `modified`.
        if '_fields' in dct:
            if not 'created' in dct['_fields']:
                dct['_fields'].update({'created'})

            if not 'modified' in dct['_fields']:
                dct['_fields'].update({'modified'})

        return super(_ComparableFrameMeta, meta).__new__(meta, name, bases, dct)


class ComparableFrame(Frame, metaclass=_ComparableFrameMeta):
    """
    A Frame-like base class that provides support for tracking changes to
    documents.

    Some important rules for creating comparable frames:

    - Override the `__str__` method of the class to return a human friendly
      identity as this method is called when generating a sticky label for the
      class.
    - Define which fields are references and which `Frame` class they reference
      in the `_compared_refs` dictionary if you don't you'll only be able to see
      that the ID has changed there will be nothing human identifiable.
    """

    # The class that will be used to store changes
    _change_log_cls = ChangeLogEntry

    # A set of fields that should be exluded from comparisons/tracking
    _uncompared_fields = {'_id', 'created', 'modified'}

    # A map of reference fields and the frames they reference
    _compared_refs = {}

    @property
    def comparable(self):
        """Return a dictionary that can be compared"""
        document_dict = self.compare_safe(self._document)

        # Remove uncompared fields
        self._remove_keys(document_dict, self._uncompared_fields)

        # Remove any empty values
        clean_document_dict = {}
        for k, v in document_dict.items():
            if not v and not isinstance(v, (int, float)):
                continue
            clean_document_dict[k] = v

        # Convert any referenced fields to Frames
        for ref_field, ref_cls in self._compared_refs.items():
            ref = getattr(self, ref_field)
            if not ref:
                continue

            # Check for fields which contain a list of references
            if isinstance(ref, list):
                if isinstance(ref[0], Frame):
                    continue

                # Dereference the list of reference IDs (retaining the order)
                docs = ref_cls.many(In(Q._id, ref))
                docs_map = {d._id: d for d in docs}
                docs = [docs_map[id] for id in ref if id in docs_map]
                clean_document_dict[ref_field] = docs

            else:
                if isinstance(ref, Frame):
                    continue

                # Dereference the reference ID
                clean_document_dict[ref_field] = ref_cls.by_id(ref)

        return clean_document_dict

    def logged_delete(self, user):
        """Delete the document and log the event in the change log"""

        # Dete the frame's document
        self.delete()

        # Log the change
        entry = self.__class__._change_log_cls({
            'type': 'DELETED',
            'documents': [self],
            'user': user
            })
        entry.insert()

        return entry

    def logged_insert(self, user):
        """Create and insert the document and log the event in the change log"""

        # Timestamp
        Frame.timestamp_insert(self, [self])

        # Insert the frame's document
        self.insert()

        # Log the insert
        entry = self.__class__._change_log_cls({
            'type': 'ADDED',
            'documents': [self],
            'user': user
            })
        entry.insert()

        return entry

    def logged_update(self, user, data, *fields):
        """
        Update the document with the dictionary of data provided and log the
        event in the change log.
        """

        # Get a copy of the frames comparable data before the update
        original = self.comparable

        # Update the frame
        _fields = fields
        if len(fields) == 0:
             _fields = data.keys()

        for field in _fields:
            if field in data:
                setattr(self, field, data[field])

        # Timestamp
        Frame.timestamp_update(self, [self])

        self.update(*fields)

        # Create an entry and perform a diff
        entry = self.__class__._change_log_cls({
            'type': 'UPDATED',
            'documents': [self],
            'user': user
            })
        entry.add_diff(original, self.comparable)

        # Check there's a change to apply/log
        if not entry.is_diff:
            return
        entry.insert()

        return entry

    @classmethod
    def compare_safe(cls, value):
        """Return a value that can be safely compared"""

        # Date
        if type(value) == date:
            return str(value)

        # Lists
        elif isinstance(value, (list, tuple)):
            return [cls.compare_safe(v) for v in value]

        # Dictionaries
        elif isinstance(value, dict):
            return {k: cls.compare_safe(v) for k, v in value.items()}

        return value



import pytest

from manhattan.comparable.change_log import ChangeLogEntry

pytest_plugins = ['fixtures']


def test_init(mongo_client):
    """Should create a new ChangeLogEntry instance"""
    entry = ChangeLogEntry()
    assert isinstance(entry, ChangeLogEntry)

def test_diff_html(mongo_client):
    """Should return an entry's diff as HTML"""

    original = {
        'bar': 'bar',
        'foo': 'foo'
        }

    new = {
        'bar': 'foo',
        'zee': 'zee'
        }

    entry = ChangeLogEntry()
    entry.add_diff(original, new)

    assert entry.diff_html.strip() == '''<div class="change change--add">
    <div class="change__field">zee</div>
    <div class="change__values">
        <div class="change__value change__value--new">
            zee
        </div>
    </div>
</div>
<div class="change change--update">
    <div class="change__field">bar</div>
    <div class="change__values">
        <div class="change__value change__value--original">
            bar
        </div>
        <div class="change__value change__value--new">
            foo
        </div>
    </div>
</div>
<div class="change change--delete">
    <div class="change__field">foo</div>
    <div class="change__values">
        <div class="change__value change__value--original">
            foo
        </div>
    </div>
</div>'''

def test_is_diff(mongo_client):
    """Should return True if there are any logged difference for the entry"""

    entry = ChangeLogEntry()

    entry.add_note('Testing no diff')
    assert entry.is_diff is False

    entry.add_diff({'foo': 'bar'}, {})
    assert entry.is_diff is True

def test_add_diff(mongo_client):
    """
    Should set the detials of the entry as the difference between 2 dictionaries
    (original vs. new).
    """

    original = {
        'bar': 'bar',
        'foo': 'foo'
        }

    new = {
        'bar': 'foo',
        'zee': 'zee'
        }

    entry = ChangeLogEntry()
    entry.add_diff(original, new)

    assert entry.details == {
        'additions': {'zee': 'zee'},
        'deletions': {'foo': 'foo'},
        'updates': {'bar': ['bar', 'foo']}
        }

def test_add_note(mongo_client):
    """Should add a note as the entries details"""

    entry = ChangeLogEntry()
    entry.add_note('Testing')

    assert entry.details == {'note': 'Testing'}

def test_diff_to_html(mongo_client):
    """Should return the given set of details as HTML"""

    # Additions
    additions = ChangeLogEntry.diff_to_html({'additions': {'zee': 'zee'}})
    assert additions == '''<div class="change change--add">
    <div class="change__field">zee</div>
    <div class="change__values">
        <div class="change__value change__value--new">
            zee
        </div>
    </div>
</div>'''

    # Updates
    updates = ChangeLogEntry.diff_to_html({'updates': {'bar': ['bar', 'foo']}})
    assert updates == '''<div class="change change--update">
    <div class="change__field">bar</div>
    <div class="change__values">
        <div class="change__value change__value--original">
            bar
        </div>
        <div class="change__value change__value--new">
            foo
        </div>
    </div>
</div>'''

    # Deletions
    deletions = ChangeLogEntry.diff_to_html({'deletions': {'foo': 'foo'}})
    assert deletions == '''<div class="change change--delete">
    <div class="change__field">foo</div>
    <div class="change__values">
        <div class="change__value change__value--original">
            foo
        </div>
    </div>
</div>'''

    # Note
    note = ChangeLogEntry.diff_to_html({'note': 'Testing'})
    assert note == '<div class="change change--note">Testing</div>'

def test_diff_safe(mongo_client, dragon):
    """Should return a value that can safely be stored as a diff"""

    safe_dragon = {'_str': str(dragon), '_id': dragon._id}

    assert ChangeLogEntry.diff_safe('foo') == 'foo'
    assert ChangeLogEntry.diff_safe(dragon) == safe_dragon
    assert ChangeLogEntry.diff_safe([dragon]) == [safe_dragon]



from mongoframes.factory.blueprints import Blueprint as BaseBlueprint

from .frames import ComparableFrame

__all__ = ['Blueprint']


class Blueprint(BaseBlueprint):
    """
    A factory blueprint for create fixture data for comparable frames.
    """

    _frame_cls = ComparableFrame

    @classmethod
    def on_fake(cls, frames):
        """Called before frames are inserted"""

        # Timestamp
        for frame in frames:
            cls._frame_cls.timestamp_insert(frame, [frame])

        super().on_fake(frames)

    @classmethod
    def on_faked(cls, frames):
        """Called after frames are inserted"""

        # Add change log entries
        for frame in frames:
            entry = cls._frame_cls._change_log_cls({
                'type': 'ADDED',
                'documents': [frame],
                'user': None
                })
            entry.insert()

        super().on_faked(frames)
